/*
 * settings.h - Definitions of settings keys and schema.
 */

#ifndef __SETTINGS_H__
#define __SETTINGS_H__

#define NOTIFICATIONS_SCHEMA             "org.ayatana.indicator.notifications"
#define NOTIFICATIONS_KEY_BLACKLIST      "blacklist"
#define NOTIFICATIONS_KEY_CLEAR_MC       "clear-on-middle-click"
#define NOTIFICATIONS_KEY_HIDE_INDICATOR "hide-indicator"
#define NOTIFICATIONS_KEY_MAX_ITEMS      "max-items"

#endif
